-----------------------------------------------------------------------------
--
-- Per window configuration:
--  Title
--  Title appearance
--  Border
--  Background color
--  BarValue
--  SortCriteria
--  Text Tags
--  Bar appearance
--
-----------------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")
local AssessmentUserInterface = Assessment:NewModule("UserInterface", "AceEvent-2.0")

-----------------------------------------------------------------------------

AssessmentUserInterface.db = Assessment:AcquireDBNamespace("UserInterface")

local profileDefaults =
{
	windows =
	{
		count = 1,

		["**"] =
		{
			displaySetStack		= { "Damage Dealt" },
			displaySetStackLevel= 1,
			updateRate			= 5,
			title = 
			{
				enable			= true,
				alignment		= "CENTER",
				font			= "Arial Narrow",
				fontSize		= 9,
				outline			= "OUTLINE",
				shadow			= true,
				height			= 10,
				textColor		= { 1, 1, 1, 1 },
				backgroundColor	= { 0, 0, 0, 1 }
			},
			bar = 
			{ 
				count			= 10,
				texture			= "Smooth",
				classList		= { },
				defaultColor	= { 0.75, 0.75, 0.75 },
			},
			frame = 
			{ 
				width			= 110,
				height			= 140, 
				point			= "CENTER",
				relativePoint	= "CENTER",
				x				= 0, 
				y				= 0, 
				borderStyle		= "Default",
				borderColor		= { 1, 1, 1, 0.75 },
				backdropColor	= { 0, 0, 0, 0.75 },
				visible			= true,
			},
		},
	},
	displaySets =
	{
		-- Defaults are inserted here in OnInitialize
	
		["**"] =
		{
			titleTag		= "Assessment",
			
			listType		= "GroupMembers", 
			barValueTag		= "damageDealt", 
			sortKeyTag		= "barValue",
			filterType		= "NONZERO",
			descendingSort	= true,
			fightIndex		= 0,
			colorTypeList	= { "Class", count = 1 },
			
			text =
			{
				count = 2,
				{
					tag			= "[rank]. [key]",
					offsets =
					{
						right	= -20,
					},
				},
				{
					tag			= "[barValue]",
					alignment	= "RIGHT",
				},
				["**"] = 
				{
					alignment	= "LEFT",
					font		= "Arial Narrow",
					fontSize	= 9,
					outline		= "",
					shadow		= true,
					alpha		= 1,
					offsets =
					{
						left		= 0,
						right		= 0,
						vertical	= 0,
					},
				}
			},
		},
	}
}

-----------------------------------------------------------------------------

local UIBar = AceOO.Class()
function UIBar.prototype:init(parent, windowOptions, displaySet, index)
	UIBar.super.prototype.init(self)

	self.bar = parent:CreateTexture()
	self.texts = { }
	self.barColorMethods = { }

	self:SetOptions(parent, windowOptions, displaySet, index)
end

function UIBar.prototype:SetOptions(parent, windowOptions, displaySet, index)
	self.windowOptions = windowOptions
	self.displaySet = displaySet

	local bar = self.bar

	for i in ipairs(self.barColorMethods) do
		self.barColorMethods[i] = nil
	end
	for i = 1, displaySet.colorTypeList.count do
		self.barColorMethods[i] = Assessment.ColorType[displaySet.colorTypeList[i]].method
	end
	self.barColorMethods[#self.barColorMethods+1] = function() return windowOptions.bar.defaultColor end

	bar:SetTexture(SML:Fetch("statusbar", windowOptions.bar.texture))
	bar:SetVertexColor(1, 1, 1, 1)
	bar:Hide()

	for i = 1, self.displaySet.text.count do
		if not self.texts[i] then
			self.texts[i] = parent:CreateFontString()
		end
	end

	for i = self.displaySet.text.count + 1, #self.texts do
		self.texts[i]:Hide()
	end
	
	self:PlaceBar(parent, index)
end

function UIBar.prototype:PlaceBar(parent, index)
	local bar = self.bar
	local verticalOffset = -5 - (self.windowOptions.title.enable and self.windowOptions.title.height or 0)
	local height = (parent:GetHeight()+verticalOffset-5) / self.windowOptions.bar.count
	
	bar:ClearAllPoints()
	bar:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, -(index-1)*height+verticalOffset)
	bar:SetPoint("BOTTOMLEFT", parent, "TOPLEFT", 5, -index*height+verticalOffset)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		local options = self.displaySet.text[i]
		
		text:ClearAllPoints()
		text:SetPoint("LEFT", parent, "LEFT", options.offsets.left+5, 0)
		text:SetPoint("RIGHT", parent, "RIGHT", options.offsets.right-5, 0)
		text:SetPoint("TOP", parent, "TOP", 0, options.offsets.vertical-(index-1)*height+verticalOffset)
		text:SetPoint("BOTTOM", parent, "TOP", 0, options.offsets.vertical-index*height+verticalOffset)

		text:SetAlpha(options.alpha)

		text:SetFont(SML:Fetch("font", options.font), options.fontSize, options.outline)
		if options.shadow then
			text:SetShadowOffset(1, -1)
		else
			text:SetShadowOffset(0, 0)
		end
		text:SetJustifyH(options.alignment)
	end
end

function UIBar.prototype:SetFraction(fraction)
	local width = fraction * (self.windowOptions.frame.width-10)
	if width == 0 then
		self.bar:Hide()
	else
		self.bar:SetWidth(width)
		self.bar:Show()
	end
end

function UIBar.prototype:Hide()
	self.bar:Hide()
	for i = 1, self.displaySet.text.count do
		self.texts[i]:Hide()
	end
end

function UIBar.prototype:Update(data, barNormalizeMultiplier, textMethods)
	-- Update Bar
	self:SetFraction(data.barValue * barNormalizeMultiplier)
	
	local color
	for i,f in ipairs(self.barColorMethods) do
		color = f(data.key)
		if color then break end
	end
	self.bar:SetVertexColor(unpack(color))
	
	-- Update Texts
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		text:SetText(textMethods[i](data))
		text:Show()
	end
end

-----------------------------------------------------------------------------

local borderStyles =
{
	None		= { edgeFile = "", inset = 5 },
	Default		= { edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", inset = 5 },
	Textured	= { edgeFile = "", inset = 3 },
	Square		= { edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", inset = 5 },
}

local UIWindow = AceOO.Class("AceEvent-2.0")

function UIWindow.prototype:init(windowOptions, displaySet)
	UIWindow.super.prototype.init(self)
	local frame = CreateFrame("Frame", nil, UIParent)
	frame.window = self

	self.frame			= frame
	self.bars			= { }
	
	self.titleText = frame:CreateFontString()
	self.titleTexture = frame:CreateTexture()
    self.titleTexture:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")

	local bottomRightDrag = CreateFrame("Frame", nil, frame)
	bottomRightDrag:Show()
	bottomRightDrag:SetFrameLevel( frame:GetFrameLevel() + 10)
	bottomRightDrag:SetWidth(16)
	bottomRightDrag:SetHeight(16)
	bottomRightDrag:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0)
	bottomRightDrag:EnableMouse(true)
	bottomRightDrag:SetScript("OnMouseDown", function() this:GetParent():StartSizing("BOTTOMRIGHT") end )
	bottomRightDrag:SetScript("OnMouseUp", function() this:GetParent():StopMovingOrSizing() end )

	self:SetWindowOptions(windowOptions)
end

local bgTable = 
{
    bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
    tile = true,
    tileSize = 16,
    edgeSize = 16,
	insets = { left = inset, right = inset, top = inset, bottom = inset }
}

function UIWindow.prototype:SetWindowOptions(windowOptions)
	local displaySet = AssessmentUserInterface:GetDisplaySet(windowOptions)

	self.dependentData		= { }
	self.dataUpdateTimes	= { }
	self.textMethods		= { }

	self.topBarOffset	= 0
	self.forceUpdate	= true
	self.displaySet		= displaySet
	self.windowOptions	= windowOptions

	self.listTypeMethod = Assessment.ListType[displaySet.listType].method
	self.filterMethod	= Assessment.FilterType[displaySet.filterType].method

	for i=1,displaySet.colorTypeList.count do
		self:MergeRequiredData(Assessment.ColorType[displaySet.colorTypeList[i]].data)
	end
	self:MergeRequiredData(Assessment.ListType[displaySet.listType].data)

	if displaySet.descendingSort then
		self.sortDirectionMethod = function(a, b) return a.sortValue > b.sortValue end
	else
		self.sortDirectionMethod = function(a, b) return a.sortValue < b.sortValue end
	end

	local compiledBarValue = Assessment:CompileTag(displaySet.barValueTag, "number")
	self.barValueMethod = compiledBarValue.method
	self:MergeRequiredData(compiledBarValue.data)
	
	for i=1,displaySet.text.count do
		local compiledTextValue = Assessment:CompileTag(displaySet.text[i].tag, "string")
		self.textMethods[i] = compiledTextValue.method
		self:MergeRequiredData(compiledTextValue.data)
	end

	local compiledSortValue = Assessment:CompileTag(displaySet.sortKeyTag, "raw")
	self.sortMethod = compiledSortValue.method

	-------

	local frame = self.frame
	local frameOptions = windowOptions.frame
	frame:SetWidth(frameOptions.width)
	frame:SetHeight(frameOptions.height)
	frame:SetPoint(frameOptions.point, UIParent, frameOptions.relativePoint, frameOptions.x, frameOptions.y)
	frame:SetMovable(true)
	frame:SetResizable(true)
	frame:EnableMouse(true)
	frame:SetMinResize(32,32)
	frame:EnableMouseWheel(true)

	frame:SetScript("OnMouseWheel", function(this, delta) 
										local window = this.window
										if delta < 0 then
											window.topBarOffset = window.topBarOffset + 1
										else
											window.topBarOffset = window.topBarOffset - 1
											if window.topBarOffset < 0 then
												window.topBarOffset = 0
											end
										end
										window.forceUpdate = true
										window:Update()
									end)
	
	frame:SetScript("OnMouseDown",		function(this) this:StartMoving() end)
	frame:SetScript("OnMouseUp",		function(this) 
											this:StopMovingOrSizing() 
											local point,_,relativePoint,x,y = this:GetPoint(1)
											local frameOptions = this.window.windowOptions.frame
											frameOptions.point = point
											frameOptions.relativePoint = relativePoint
											frameOptions.x = x
											frameOptions.y = y
										end)
	frame:SetScript("OnSizeChanged",	function(this, width, height) 
											local window = this.window
											local frameOptions = window.windowOptions.frame
											frameOptions.width = width 
											frameOptions.height = height
											window:LayoutBars()
											window.forceUpdate = true
											window:Update()
										end)


	local borderStyle = borderStyles[windowOptions.frame.borderStyle]
	bgTable.edgeFile = borderStyle.edgeFile
	local inset = borderStyle.inset
	local insets = bgTable.insets
	insets.left = inset
	insets.top = inset
	insets.right = inset
	insets.bottom = inset
    frame:SetBackdrop(bgTable)
	frame:SetBackdropBorderColor(unpack(windowOptions.frame.borderColor))
	frame:SetBackdropColor(unpack(windowOptions.frame.backdropColor))

	--------

	local titleText = self.titleText
	local titleOptions = windowOptions.title
	
	if titleOptions.enable then
		titleText:ClearAllPoints()
		titleText:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleText:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleText:SetTextColor(unpack(titleOptions.textColor))

		titleText:SetFont(SML:Fetch("font", titleOptions.font), titleOptions.fontSize, titleOptions.outline)
		if titleOptions.shadow then
			titleText:SetShadowOffset(1, -1)
		else
			titleText:SetShadowOffset(0, 0)
		end
		titleText:SetJustifyH(titleOptions.alignment)

		local compiledTitleTextValue = Assessment:CompileTag(displaySet.titleTag, "string")
		self.titleTextMethod = compiledTitleTextValue.method
		self:MergeRequiredData(compiledTitleTextValue.data)

		titleText:Show()

		local titleTexture = self.titleTexture
		titleTexture:ClearAllPoints()
		titleTexture:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleTexture:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleTexture:SetVertexColor(unpack(titleOptions.backgroundColor))
		titleTexture:Show()
	else
		titleText:Hide()
		self.titleTexture:Hide()
	end

	--------

	for i = 1, windowOptions.bar.count do
		if self.bars[i] then
			self.bars[i]:SetOptions(frame, windowOptions, displaySet, i)
		else
			self.bars[i] = UIBar:new(frame, windowOptions, displaySet, i)
		end
	end
	for i = windowOptions.bar.count+1, #self.bars do
		self.bars[i]:Hide()
	end

	if frameOptions.visible then
		self:Show()
	else
		self:Hide()
	end
end

function UIWindow.prototype:Show()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
	end

	local windowOptions = self.windowOptions
	local updateRate = windowOptions.updateRate ~= 0 and 1/windowOptions.updateRate or 0

	self.updater = self:ScheduleRepeatingEvent(self.Update, updateRate, self)
	self.frame:Show()
	self:Update()
end

function UIWindow.prototype:Hide()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
		self.updater = nil
	end
	self.frame:Hide()
end

function UIWindow.prototype:LayoutBars(fightData)
	for i=1, self.windowOptions.bar.count do
		self.bars[i]:PlaceBar(self.frame, i)
	end
end

function UIWindow.prototype:MergeRequiredData(data)
	if not data then return end
	if type(data) == "string" then
		self.dependentData[data] = true
	elseif type(data) == "table" then
		for k in pairs(data) do
			self.dependentData[k] = true
		end
	end
end

function UIWindow.prototype:NeedsUpdate(fightData)
	local dataUpdateTimes = self.dataUpdateTimes

	if not fightData then
		if next(dataUpdateTimes) then
			for k in pairs(dataUpdateTimes) do
				dataUpdateTimes[k] = nil
			end
			return true
		end
		if self.forceUpdate then
			self.forceUpdate = nil
			return true
		end
		return false
	end

	local result = false
	for k in pairs(self.dependentData) do
		if dataUpdateTimes[k] ~= fightData.dataUpdateTimes[k] then
			result = true
			dataUpdateTimes[k] = fightData.dataUpdateTimes[k]
		end
	end
	if self.forceUpdate then
		self.forceUpdate = nil
		return true
	end
	return result
end

local sortList = { }
local titleData = { key = UnitName("player"), rank = 0 }
function UIWindow.prototype:DoUpdate(fightData)
	local maximumBarValue = 0
	local totalBarValue = 0

	for i,v in ipairs(sortList) do
		sortList[i] = Assessment:ReleaseTable(v)
	end

	if fightData then
		for k,v in self.listTypeMethod(fightData) do
			local data = Assessment:NewTable()
			data.key = k
			data.value = v
			data.fightData = fightData
			
			local barValue = self.barValueMethod(data)
			data.barValue = barValue
			
			if self.filterMethod(data) then
				if barValue > maximumBarValue then
					maximumBarValue = barValue
				end
				totalBarValue = totalBarValue + barValue

				data.barValue = barValue
				sortList[#sortList+1] = data
			else
				Assessment:ReleaseTable(data)
			end			
		end
		
		for _,data in ipairs(sortList) do
			data.totalBarValue = totalBarValue
			data.maximumBarValue = maximumBarValue
			data.sortValue = self.sortMethod(data)
		end

		table.sort(sortList, self.sortDirectionMethod)
	end

	local barNormalizeMultiplier = maximumBarValue ~= 0 and 1 / maximumBarValue or 0

	local topBarOffset = self.topBarOffset
	local barCount = self.windowOptions.bar.count
	if topBarOffset + barCount > #sortList then
		topBarOffset = #sortList - barCount
		if topBarOffset < 0 then
			topBarOffset = 0
		end
		self.topBarOffset = topBarOffset
	end

	local endDisplay = barCount
	if endDisplay > #sortList-topBarOffset then 
		endDisplay = #sortList-topBarOffset 
		for i = endDisplay+1, barCount do
			self.bars[i]:Hide()
		end
	end
	for i = 1, endDisplay do
		local rank = i+topBarOffset
		local data = sortList[rank]
		data.rank = rank
		self.bars[i]:Update(data, barNormalizeMultiplier, self.textMethods)
	end	

	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue

		self.titleText:SetText( self.titleTextMethod(titleData) )
	end
end

function UIWindow.prototype:Update()
	local fightData = Assessment:GetFightData(self.displaySet.fightIndex)
	if not self:NeedsUpdate(fightData) then return end
	self:DoUpdate(fightData)
end

-----------------------------------------------------------------------------

function AssessmentUserInterface:OnInitialize()
	self.windows = { }

	for k,v in pairs(Assessment.DefaultDisplaySets) do
		local displaySet = { }
		profileDefaults.displaySets[k] = displaySet
		displaySet.titleTag		= v.titleTag
		displaySet.listType		= v.listType
		displaySet.barValueTag	= v.barValueTag
		displaySet.filterType	= v.filterType
		displaySet.fightIndex	= v.fightIndex
		if v.barValueTextTag then
			displaySet.text = { [2] = { tag = v.barValueTextTag }	}
		end
	end
	
	Assessment:RegisterDefaults("UserInterface", "profile", profileDefaults)
	
	self:CreateValidateLists()
end

function AssessmentUserInterface:OnEnable()
	self:RegisterEvent('SharedMedia_SetGlobal', function(ty, val)
		if ty == "statusbar" then
			for k,v in ipairs(self.windows) do
				for k,v in ipairs(v.bars) do
					v.bar:SetTexture(SML:Fetch(ty, val))
				end
			end
		end
	end)
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentUserInterface:OnDisable()
end

function AssessmentUserInterface:CreateWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self.windows[i] then
			self.windows[i]:SetWindowOptions(windowOptions)
		else
			self.windows[i] = UIWindow:new(windowOptions)
		end
	end
	
	for i = self.db.profile.windows.count+1, #self.windows do
		self.windows[i]:Hide()
	end
end

function AssessmentUserInterface:GetDisplaySet(windowOptions)
	return self.db.profile.displaySets[windowOptions.displaySetStack[windowOptions.displaySetStackLevel]]
end

-----------------------------------------------------------------------------
